from functools import lru_cache

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import leastsq
from scipy.optimize import curve_fit
from scipy.signal import fftconvolve

import toolbox_scs as tb


__all__ = [
    'find_curvature',
    'correct_curvature',
    'get_spectrum',
    'energy_calibration',
    'calibrate',
    'gaussian_fit',
    'to_fwhm'
]


# -----------------------------------------------------------------------------
# Curvature

def find_curvature(image, frangex=None, frangey=None,
                   deg=2, axis=1, offset=100):
    # Resolve arguments
    x_range = (0, image.shape[1])
    if frangex is not None:
        x_range = (max(frangex[0], x_range[0]), min(frangex[1], x_range[1]))
    y_range = (0, image.shape[0])
    if frangex is not None:
        y_range = (max(frangey[0], y_range[0]), min(frangey[1], y_range[1]))

    axis_range = y_range if axis == 1 else x_range
    axis_dim = image.shape[axis - 1]

    # Get kernel
    integral = image[slice(*y_range), slice(*x_range)].mean(axis=axis)
    roi = np.ones([axis_range[1] - axis_range[0], axis_dim])
    ref = roi * integral[:, np.newaxis]

    # Get sliced image
    slice_ = [slice(None), slice(None)]
    slice_[axis - 1] = slice(max(axis_range[0] - offset, 0),
                             min(axis_range[1] + offset, axis_dim))
    sliced = image[tuple(slice_)]
    if axis == 0:
        sliced = sliced.T

    # Get curvature factor from cross correlation
    crosscorr = fftconvolve(sliced,
                            ref[::-1, :],
                            axes=0, )
    shifts = np.argmax(crosscorr, axis=0)
    curv = np.polyfit(np.arange(axis_dim), shifts, deg=deg)
    return curv[:-1][::-1]


def find_curvature(img, args, plot=False, **kwargs):
    def parabola(x, a, b, c, s=0, h=0, o=0):
        return (a*x + b)*x + c
    def gauss(y, x, a, b, c, s, h, o=0):
        return h * np.exp(-((y - parabola(x, a, b, c)) / (2 * s))**2) + o
    x = np.arange(img.shape[1])[None, :]
    y = np.arange(img.shape[0])[:, None]

    if plot:
        plt.figure(figsize=(10,10))
        plt.imshow(img, cmap='gray', aspect='auto', interpolation='nearest', **kwargs)
        plt.plot(x[0, :], parabola(x[0, :], *args))

    args, _ = leastsq(lambda args: (gauss(y, x, *args) - img).ravel(), args)

    if plot:
        plt.plot(x[0, :], parabola(x[0, :], *args))
    return args


def correct_curvature(image, factor=None, axis=1):
    if factor is None:
        return

    if axis == 1:
        image = image.T

    ydim, xdim = image.shape
    x = np.arange(xdim + 1)
    y = np.arange(ydim + 1)
    xx, yy = np.meshgrid(x[:-1] + 0.5, y[:-1] + 0.5)
    xxn = xx - factor[0] * yy - factor[1] * yy ** 2
    ret = np.histogramdd((xxn.flatten(), yy.flatten()),
                         bins=[x, y],
                         weights=image.flatten())[0]

    return ret if axis == 1 else ret.T


def get_spectrum(image, factor=None, axis=0,
                 pixel_range=None, energy_range=None, ):
    start, stop = (0, image.shape[axis - 1])
    if pixel_range is not None:
        start = max(pixel_range[0] or start, start)
        stop = min(pixel_range[1] or stop, stop)

    edge = image.sum(axis=axis)[start:stop]
    bins = np.arange(start, stop + 1)
    centers = (bins[1:] + bins[:-1]) * 0.5
    if factor is not None:
        centers, edge = calibrate(centers, edge,
                                  factor=factor,
                                  range_=energy_range)

    return centers, edge


# -----------------------------------------------------------------------------
# Energy calibration


def energy_calibration(channels, energies):
    return np.polyfit(channels, energies, deg=1)


def calibrate(x, y=None, factor=None, range_=None):
    if factor is not None:
        x = np.polyval(factor, x)

    if y is not None and range_ is not None:
        start = np.argmin(np.abs((x - range_[0])))
        stop = np.argmin(np.abs((x - range_[1])))
        # Calibrated energies have a different direction
        x, y = x[stop:start], y[stop:start]

    return x, y


# -----------------------------------------------------------------------------
# Gaussian-related functions


FWHM_COEFF = 2 * np.sqrt(2 * np.log(2))


def gaussian_fit(x_data, y_data, offset=0):
    """
    Centre-of-mass and width. Lifted from image_processing.imageCentreofMass()
    """

    x0 = np.average(x_data, weights=y_data)
    sx = np.sqrt(np.average((x_data - x0) ** 2, weights=y_data))

    # Gaussian fit
    baseline = y_data.min()
    p_0 = (y_data.max(), x0 + offset, sx, baseline)
    try:
        p_f, _ = curve_fit(gauss1d, x_data, y_data, p_0, maxfev=10000)
        return p_f
    except (RuntimeError, TypeError) as e:
        print(e)
        return None


def gauss1d(x, height, x0, sigma, offset):
    return height * np.exp(-0.5 * ((x - x0) / sigma) ** 2) + offset


def to_fwhm(sigma):
    return abs(sigma * FWHM_COEFF)


# -----------------------------------------------------------------------------
# Centroid

THRESHOLD = 510  # pixel counts above which a hit candidate is assumed
CURVE_A = 2.19042931e-02  # curvature parameters as determined elsewhere
CURVE_B = -3.02191568e-07


def _esrf_centroid(image, threshold=THRESHOLD, curvature=(CURVE_A, CURVE_B)):
    gs = 2
    base = image.mean()
    cp = np.argwhere(image[gs // 2: -gs // 2, gs // 2: -gs // 2] > threshold) + np.array([gs // 2, gs // 2])
    if len(cp) > 100000:
        raise RuntimeError('Threshold too low or acquisition time too long')

    res = []
    for cy, cx in cp:
        spot = image[cy - gs // 2: cy + gs // 2 + 1, cx - gs // 2: cx + gs // 2 + 1] - base
        spot[spot < 0] = 0
        if (spot > image[cy, cx]).sum() == 0:
            mx = np.average(np.arange(cx - gs // 2, cx + gs // 2 + 1), weights=spot.sum(axis=0))
            my = np.average(np.arange(cy - gs // 2, cy + gs // 2 + 1), weights=spot.sum(axis=1))
            my -= (curvature[0] + curvature[1] * mx) * mx
            res.append((my, mx))
    return res


def _new_centroid(image, threshold=THRESHOLD, std_threshold=3.5, curvature=(CURVE_A, CURVE_B)):
    """find the position of photons with sub-pixel precision

    A photon is supposed to have hit the detector if the intensity within a
    2-by-2 square exceeds a threshold. In this case the position of the photon
    is calculated as the center-of-mass in a 4-by-4 square.

    Return the list of x,y coordinate pairs, corrected by the curvature.
    """
    base = image.mean()
    corners = image[1:, 1:] + image[:-1, 1:] + image[1:, :-1] + image[:-1, :-1]
    if threshold is None:
        threshold = corners.mean() + std_threshold * corners.std()
    middle = corners[1:-1, 1:-1]
    candidates = (
            (middle > threshold)
            * (middle >= corners[:-2, 1:-1]) * (middle > corners[2:, 1:-1])
            * (middle >= corners[1:-1, :-2]) * (middle > corners[1:-1, 2:])
            * (middle >= corners[:-2, :-2]) * (middle > corners[2:, :-2])
            * (middle >= corners[:-2, 2:]) * (middle > corners[2:, 2:]))
    cp = np.argwhere(candidates)
    if len(cp) > 10000:
        raise RuntimeError(
            "too many peaks, threshold too low or acquisition time too high")

    res = []
    for cy, cx in cp:
        spot = image[cy: cy + 4, cx: cx + 4] - base
        mx = np.average(np.arange(cx, cx + 4), weights=spot.sum(axis=0))
        my = np.average(np.arange(cy, cy + 4), weights=spot.sum(axis=1))
        my -= (curvature[0] + curvature[1] * mx) * mx
        res.append((my, mx))
    return res


centroid = _new_centroid


def decentroid(res):
    res = np.array(res)
    ret = np.zeros(shape=(res.max(axis=0) + 1).astype(int))
    for cy, cx in res:
        if cx > 0 and cy > 0:
            ret[int(cy), int(cx)] += 1
    return ret


# -----------------------------------------------------------------------------
# Integral

FACTOR = 3
RANGE = [300, 400]
BINS = abs(np.subtract(*RANGE)) * FACTOR


def parabola(x, a, b, c=0):
    return (a * x + b) * x + c


def integrate(image, factor=FACTOR, range=RANGE, curvature=(CURVE_A, CURVE_B), ):
    image = image - image.mean()
    x = np.arange(image.shape[1])[None, :]
    y = np.arange(image.shape[0])[:, None]
    ys = factor * (y - parabola(x, curvature[1], curvature[0]))
    ysf = np.floor(ys)
    rang = (factor * range[0], factor * range[1])
    bins = rang[1] - rang[0]
    lhy, lhx = np.histogram(ysf.ravel(), bins=bins, weights=((ys - ysf) * image).ravel(), range=rang)
    rhy, rhx = np.histogram((ysf + 1).ravel(), bins=bins, weights=(((ysf + 1) - ys) * image).ravel(), range=rang)
    lvy, lvx = np.histogram(ysf.ravel(), bins=bins, weights=(ys - ysf).ravel(), range=rang)
    rvy, rvx = np.histogram((ysf + 1).ravel(), bins=bins, weights=((ysf + 1) - ys).ravel(), range=rang)
    return (lhy + rhy) / (lvy + rvy)


# -----------------------------------------------------------------------------
# hRIXS class


class hRIXS:
    # run
    PROPOSAL = 2769

    # image range
    X_RANGE = np.s_[:]
    Y_RANGE = np.s_[:]

    # centroid
    THRESHOLD = None  # pixel counts above which a hit candidate is assumed
    STD_THRESHOLD = 3.5  # same as THRESHOLD, in standard deviations
    CURVE_A = CURVE_A  # curvature parameters as determined elsewhere
    CURVE_B = CURVE_B

    # integral
    FACTOR = FACTOR
    RANGE = RANGE
    BINS = abs(np.subtract(*RANGE)) * FACTOR

    METHOD = 'centroid'  # ['centroid', 'integral']
    USE_DARK = False

    ENERGY_INTERCEPT = 0
    ENERGY_SLOPE = 1

    FIELDS = ['hRIXS_det', 'hRIXS_index', 'hRIXS_delay', 'hRIXS_norm']

    def set_params(self, **params):
        for key, value in params.items():
            setattr(self, key.upper(), value)

    def get_params(self, *params):
        if not params:
            params = ('proposal', 'x_range', 'y_range',
                      'threshold', 'curve_a', 'curve_b',
                      'factor', 'range', 'bins',
                      'method', 'fields')
        return {param: getattr(self, param.upper()) for param in params}

    def from_run(self, runNB, proposal=None, extra_fields=()):
        if proposal is None:
            proposal = self.PROPOSAL
        run, data = tb.load(proposal, runNB=runNB,
                            fields=self.FIELDS + list(extra_fields))

        return data

    def load_dark(self, runNB, proposal=None):
        data = self.from_run(runNB, proposal)
        self.dark_image = data['hRIXS_det'].mean(dim='trainId')
        self.USE_DARK = True

    def find_curvature(self, runNB, proposal=None, plot=True, args=None, **kwargs):
        data = self.from_run(runNB, proposal)

        image = data['hRIXS_det'].sum(dim='trainId') \
                .to_numpy()[self.X_RANGE, self.Y_RANGE].T
        if args is None:
            spec = (image - image[:10, :].mean()).mean(axis=1)
            mean = np.average(np.arange(len(spec)), weights=spec)
            args = (-2e-7, 0.02, mean - 0.02 * image.shape[1] / 2, 3,
                    spec.max(), image.mean())
        args = find_curvature(image, args, plot=plot, **kwargs)
        self.CURVE_B, self.CURVE_A, *_ = args
        return self.CURVE_A, self.CURVE_B

    def centroid(self, data, bins=None):
        if bins is None:
            bins = self.BINS
        ret = np.zeros((len(data["hRIXS_det"]), bins))
        for image, r in zip(data["hRIXS_det"], ret):
            c = centroid(
                image.to_numpy()[self.X_RANGE, self.Y_RANGE].T,
                threshold=self.THRESHOLD,
                std_threshold=self.STD_THRESHOLD,
                curvature=(self.CURVE_A, self.CURVE_B))
            if not len(c):
                continue
            rc = np.array(c)
            hy, hx = np.histogram(
                rc[:, 0], bins=bins,
                range=(0, self.Y_RANGE.stop - self.Y_RANGE.start))
            r[:] = hy

        data = data.assign_coords(
            energy=np.linspace(self.Y_RANGE.start, self.Y_RANGE.stop, bins)
            * self.ENERGY_SLOPE + self.ENERGY_INTERCEPT)
        return data.assign(spectrum=(("trainId", "energy"), ret))

    def integrate(self, data):
        bins = self.Y_RANGE.stop - self.Y_RANGE.start
        ret = np.zeros((len(data["hRIXS_det"]), bins - 20))
        for image, r in zip(data["hRIXS_det"], ret):
            if self.USE_DARK:
                image = image - self.dark_image
            r[:] = integrate(image.to_numpy()[self.X_RANGE, self.Y_RANGE].T, factor=1,
                             range=(10, bins - 10),
                             curvature=(self.CURVE_A, self.CURVE_B))
        data = data.assign_coords(
            energy=np.arange(self.Y_RANGE.start + 10, self.Y_RANGE.stop - 10)
            * self.ENERGY_SLOPE + self.ENERGY_INTERCEPT)
        return data.assign(spectrum=(("trainId", "energy"), ret))

    aggregators = dict(
        hRIXS_det=lambda x, dim: x.sum(dim=dim),
        Delay=lambda x, dim: x.mean(dim=dim),
        hRIXS_delay=lambda x, dim: x.mean(dim=dim),
        hRIXS_norm=lambda x, dim: x.sum(dim=dim),
        spectrum=lambda x, dim: x.sum(dim=dim),
    )

    def aggregator(self, da, dim):
        agg = self.aggregators.get(da.name)
        if agg is None:
            return None
        return agg(da, dim=dim)

    def aggregate(self, ds, dim="trainId"):
        ret = ds.map(self.aggregator, dim=dim)
        ret = ret.drop_vars([n for n in ret if n not in self.aggregators])
        return ret

    def normalize(self, data):
        return data.assign(normalized=data["spectrum"] / data["hRIXS_norm"])
